import pandas as pd
import PIL
import os

import torch
from torch.utils.data import Dataset, DataLoader


class TimeSeriesDataset(Dataset):
    """
    Time Series Dataset for PyTorch RNN

    == Arguments ==
    time_series: pandas DataFrame
        the time series data

    target_col_name: string
        target column name in the DataFrame

    seq_len: int
        sequence length. How far would the model backpropagate when training.

    batch_size: int
        the size when using time series batching

    n_feature: int
        number of input features

    report: bool
        print the batched dataset information
    """
    def __init__(self, time_series, target_col_name, seq_len, batch_size, n_feature, report=True):
        self.seq_len = seq_len
        self.batch_size = batch_size
        self.n_feature = n_feature

        # Calculate maximum step
        n_step = (len(time_series) - 1) // (seq_len * batch_size)
        self.n_step = n_step

        if report:
            print(f"Using \x1b[31m{seq_len} previous data\x1b[0m as context with \x1b[31mbatching " \
                  f"of {batch_size}\x1b[0m. There would be \x1b[31m{n_step} complete time steps\x1b[0m")

        # Trim series to complete steps
        input_series = time_series[:n_step * seq_len * batch_size]
        target_series = time_series[target_col_name][1:n_step * seq_len * batch_size + 1]

        if hasattr(time_series, 'index'):
            self.input_ticks = input_series.index
            self.target_ticks = target_series.index

        # Data engineering for RNN
        batched_input_series = input_series.values.reshape(batch_size, n_step, seq_len, n_feature)
        self.X = torch.FloatTensor(batched_input_series).transpose(0, 1)

        batched_target_series = target_series.values.reshape(batch_size, n_step, seq_len, 1)
        self.y = torch.FloatTensor(batched_target_series).transpose(0, 1)

    def __getitem__(self, i):
        return self.X[i], self.y[i]

    def __len__(self):
        return len(self.X)


class TimeSeriesDataloader(DataLoader):
    """
    Time Series Dataloader when to complement jcopdl's TimeSeriesDataset

    == Arguments ==
    dataset: jcopdl.utils.TimeSeriesDataset
        the time series dataset object

    num_workers: int
        how many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.
    """
    def __init__(self, dataset, num_workers=0):
        super().__init__(dataset, batch_size=1, num_workers=num_workers, collate_fn=lambda x: x[0])


class MultilabelDataset(Dataset):
    """
    Multilabel Dataset for PyTorch
    The folder structur should be
        _________________________________
        data/
            train/
                img1.jpg
                2.jpg
                three.jpg
                ... all train images ...
            test/
                a.jpg
                b.jpg
                c.jpg
                ... all test images ...
            train_label.csv
            test_label.csv
        _________________________________

    train_label.csv and test_label.csv are the metadata. It should contain file name and its corresponding labels
    For example:
        __________________________
        fname,label1,label2,label3
        img1.jpg,1,1,0
        2.jpg,1,0,0
        three.jpg,1,1,1
        ...
        __________________________

    == Arguments ==
    csv_path: string
        the metadata file

    csv_path: string
        the img folder

    transform: torchvision.transform
        torchvision data augmentation

    fname_col: string
        header of file name used in the metadata
    """
    def __init__(self, csv_path, img_path, transform=None, fname_col='fname'):
        df = pd.read_csv(csv_path)
        assert df[fname_col].apply(lambda x: os.path.isfile(img_path + x)).all

        self.classes = df.columns[1:].to_list()

        self.img_path = img_path
        self.transform = transform

        self.X = df[fname_col]
        self.y = df.drop(columns=fname_col)

    def __getitem__(self, index):
        img = PIL.Image.open(self.img_path + self.X[index])
        if self.transform is not None:
            img = self.transform(img)

        label = torch.from_numpy(self.y.iloc[index].values).type(torch.FloatTensor)
        return img, label

    def __len__(self):
        return len(self.X)



from torch import nn

_act_func = {
    "relu": nn.ReLU(),
    "lrelu": nn.LeakyReLU(),
    "sigmoid": nn.Sigmoid(),
    "tanh": nn.Tanh(),
    "elu": nn.ELU(),
    "selu": nn.SELU(),
    "softmax": nn.Softmax(1),
    "lsoftmax": nn.LogSoftmax(1)
}


def linear_block(n_in, n_out, batch_norm=False, activation='relu', dropout=0.):
    layers = [nn.Linear(n_in, n_out)]

    if batch_norm:
        layers.append(nn.BatchNorm1d(n_out))

    if activation in _act_func:
        layers.append(_act_func[activation])
    else:
        raise Exception(f"jcopdl supports ({', '.join(_act_func.keys())})")

    if 0 < dropout <= 1:
        layers.append(nn.Dropout(dropout))
    return nn.Sequential(*layers)


def conv_block(c_in, c_out, kernel=3, stride=1, pad=1, pool_type='max', pool_kernel=2, pool_stride=2,
               batch_norm=False, activation='relu'):
    layers = [nn.Conv2d(c_in, c_out, kernel_size=kernel, stride=stride, padding=pad)]

    if batch_norm:
        layers.append(nn.BatchNorm2d(c_out))

    if activation in _act_func:
        layers.append(_act_func[activation])
    else:
        raise Exception(f"jcopdl supports ({', '.join(_act_func.keys())})")

    if pool_type == "max":
        layers.append(nn.MaxPool2d(pool_kernel, pool_stride))
    elif pool_type == "avg":
        layers.append(nn.AvgPool2d(pool_kernel, pool_stride))
    elif pool_type is None:
        pass
    else:
        raise Exception("jcopdl supports (max, avg)")
    return nn.Sequential(*layers)


def rnn_block(input_size, hidden_size, num_layers, cell_type, dropout=0, bidirectional=False):
    params = {
        "input_size": input_size,
        "hidden_size": hidden_size,
        "num_layers": num_layers,
        "dropout": dropout,
        "bidirectional": bidirectional,
        "batch_first": True
    }

    if cell_type == "rnn":
        return nn.RNN(**params)
    elif cell_type == "lstm":
        return nn.LSTM(**params)
    elif cell_type == "gru":
        return nn.GRU(**params)
    else:
        raise Exception("jcopdl supports (rnn, lstm, gru)")



class Config:
    def __init__(self):
        pass

    def __repr__(self):
        params = [f"{k}={v}" for k, v in sorted(self.__dict__.items())]
        return f"Config({', '.join(params)})"


def set_config(config_dict):
    config = Config()
    config.__dict__ = config_dict
    return config



from setuptools import setup, find_packages
import jcopdl

with open("README.md", "r") as f:
    description = f.read()

setup(
    name="jcopdl",
    version=jcopdl.__version__,
    author="Wira Dharma Kencana Putra",
    author_email="wiradharma_kencanaputra@yahoo.com",
    description="J.COp DL is a deep Learning package to complement pytorch workflow",
    long_description=description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6.7",
    install_requires=['torch', 'numpy', 'pandas', 'matplotlib', 'pillow'],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Natural Language :: Indonesian",
        "Natural Language :: English"
    ],
    keywords="deep learning dl jcop indonesia",
    url="https://github.com/WiraDKP/jcopdl"
)


import torch


def accuracy_binary(model, X, y):
    """
    Function to calculate binary classification accuracy.
    Note: Sigmoid is the final activation

    == Arguments ==
    model: torch.nn.Module
        A deep learning architecture using PyTorch nn.Module

    X: torch.Tensor
        input data

    y: torch.Tensor
        target data
    """
    with torch.no_grad():
        model.eval()
        output = model(X)

    pred = output.round().squeeze()
    acc = (y == pred).to(torch.float32).mean()
    return acc.item()


def accuracy_multiclass(model, X, y):
    """
    Function to calculate multiclass classification accuracy.
    Note: Softmax or LogSoftmax is the final activation

    == Arguments ==
    model: torch.nn.Module
        A deep learning architecture using PyTorch nn.Module

    X: torch.Tensor
        input data

    y: torch.Tensor
        target data
    """
    with torch.no_grad():
        model.eval()
        output = model(X)

    pred = output.argmax(1)
    acc = (y == pred).to(torch.float32).mean()
    return acc.item()
